#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# Attempt to download and build PDFium from source. This may take very long.
# Only tested on Linux with glibc. Might not work on other platforms.

import os
import sys
import shutil
import argparse
import urllib.request
from os.path import join, abspath, dirname

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import (
    SB_Dir,
    Libnames,
    DataTree,
    PlatformNames,
    run_cmd,
    call_ctypesgen,
)


PatchDir       = join(SB_Dir,'patches')
DepotToolsDir  = join(SB_Dir,'depot_tools')
PDFiumDir      = join(SB_Dir,'pdfium')
PDFiumBuildDir = join(PDFiumDir,'out','Default')
OutputDir      = join(DataTree,PlatformNames.sourcebuild)
NB_BinaryDir   = join(PDFiumDir,'third_party','llvm-build','Release+Asserts','bin')

DepotTools_URL = "https://chromium.googlesource.com/chromium/tools/depot_tools.git"
PDFium_URL     = "https://pdfium.googlesource.com/pdfium.git"

DepotPatches = [
    (join(PatchDir,'depot_tools','gclient_scm.patch'), DepotToolsDir),
]
PdfiumMainPatches = [
    (join(PatchDir,'pdfium','public_headers.patch'), PDFiumDir),
    (join(PatchDir,'pdfium','shared_library.patch'), PDFiumDir),
]
PdfiumWinPatches = [
    (join(PatchDir,'pdfium','win','pdfium.patch'), PDFiumDir),
    (join(PatchDir,'pdfium','win','build.patch'), join(PDFiumDir,'build')),
]

DefaultConfig = {
    'is_debug': False,
    'treat_warnings_as_errors': False,
    #'clang_use_chrome_plugins': False,
    'pdf_is_standalone': True,
    'pdf_enable_v8': False,
    'pdf_enable_xfa': False,
    'pdf_use_skia': False,
}
SyslibsConfig = {
    'use_system_freetype': True,
    'use_system_lcms2': True,
    #'use_system_libjpeg': True,
    'use_system_libopenjpeg2': True,
    'use_system_libpng': True,
    'use_system_zlib': True,
    'sysroot': '/',
}

if sys.platform.startswith('linux'):
    SyslibsConfig['use_custom_libcxx'] = False
elif sys.platform.startswith('darwin'):
    DefaultConfig['mac_deployment_target'] = '10.11.0'
    SyslibsConfig['use_system_xcode'] = True
elif sys.platform.startswith('win32'):
    DefaultConfig['pdf_use_win32_gdi'] = True


def dl_depottools(do_update):
    
    if not os.path.isdir(SB_Dir):
        os.makedirs(SB_Dir)
    
    is_update = True
    
    if os.path.isdir(DepotToolsDir):
        if do_update:
            print("DepotTools: Revert and update ...")
            run_cmd(['git', 'reset', '--hard', 'HEAD'], cwd=DepotToolsDir)
            run_cmd(['git', 'pull', DepotTools_URL], cwd=DepotToolsDir)
        else:
            print("DepotTools: Using existing repository as-is.")
            is_update = False
    else:
        print("DepotTools: Download ...")
        run_cmd(['git', 'clone', '--depth', '1', DepotTools_URL, DepotToolsDir], cwd=SB_Dir)
    
    os.environ['PATH'] += os.pathsep + DepotToolsDir
    
    return is_update


def _dl_unbundler():
    
    # Download missing tools for unbundle/replace_gn_files.py
    
    tool_url = "https://raw.githubusercontent.com/chromium/chromium/main/tools/generate_shim_headers/generate_shim_headers.py"
    tool_dir = join(PDFiumDir,"tools","generate_shim_headers")
    tool_file = join(tool_dir, "generate_shim_headers.py")
    
    if not os.path.isdir(tool_dir):
        os.makedirs(tool_dir)
    if not os.path.exists(tool_file):
        urllib.request.urlretrieve(tool_url, tool_file)


def dl_pdfium(do_update, revision, GClient):
    
    is_sync = True
    
    if os.path.isdir(PDFiumDir):
        if do_update:
            print("PDFium: Revert / Sync  ...")
            run_cmd([GClient, 'revert'], cwd=SB_Dir)
        else:
            is_sync = False
            print("PDFium: Using existing repository as-is.")
    else:
        print("PDFium: Download ...")
        run_cmd([GClient, 'config', '--unmanaged', PDFium_URL], cwd=SB_Dir)
    
    if is_sync:
        run_cmd([GClient, 'sync', '--revision', 'origin/%s' % revision, '--no-history', '--shallow'], cwd=SB_Dir)
        _dl_unbundler()
    
    return is_sync


def _apply_patchset(patchset):
    for patch, cwd in patchset:
        run_cmd(['git', 'apply', '-v', patch], cwd=cwd)

def patch_depottools():
    _apply_patchset(DepotPatches)

def patch_pdfium():
    _apply_patchset(PdfiumMainPatches)
    if sys.platform.startswith('win32'):
        _apply_patchset(PdfiumWinPatches)
        shutil.copy(join(PatchDir,'pdfium','win','resources.rc'), join(PDFiumDir,'resources.rc'))


def configure(config, GN):
    if not os.path.exists(PDFiumBuildDir):
        os.makedirs(PDFiumBuildDir)
    with open(join(PDFiumBuildDir,'args.gn'), 'w') as args_handle:
        args_handle.write(config)
    run_cmd([GN, 'gen', PDFiumBuildDir], cwd=PDFiumDir)


def build(Ninja):
    run_cmd([Ninja, '-C', PDFiumBuildDir, 'pdfium'], cwd=PDFiumDir)


def find_lib(srcname=None, directory=PDFiumBuildDir):
    
    if srcname is not None:
        path = join(PDFiumBuildDir, srcname)
        if os.path.isfile(path):
            return path
        else:
            print("Warning: The file of given srcname does not exist.", file=sys.stderr)
    
    libpath = None
    for lname in Libnames:
        path = join(directory, lname)
        if os.path.isfile(path):
            libpath = path
    
    if libpath is None:
        raise RuntimeError("Build artifact not found.")
    
    return libpath


def pack(src_libpath, destname=None):
    
    if os.path.isdir(OutputDir):
        shutil.rmtree(OutputDir)
    os.makedirs(OutputDir)
    
    if destname is None:
        destname = os.path.basename(src_libpath)
    
    destpath = join(OutputDir, destname)
    shutil.copy(src_libpath, destpath)
    
    include_dir = join(OutputDir,'include')
    shutil.copytree(join(PDFiumDir,'public'), include_dir)
    
    call_ctypesgen(OutputDir, include_dir)
    shutil.rmtree(include_dir)


def _get_tool(tool, win_append):
    exe = join(DepotToolsDir, tool)
    if sys.platform.startswith('win32'):
        exe += '.' + win_append
    return exe


def _serialise_config(config_dict):
    
    config_str = ''
    sep = ''
    
    for key, value in config_dict.items():
        config_str += sep + '%s = ' % key
        if isinstance(value, bool):
            config_str += str(value).lower()
        elif isinstance(value, str):
            config_str += '"%s"' % value
        else:
            raise TypeError("Not sure how to serialise type %s" % type(value))
        sep = '\n'
    
    return config_str


def main(
        b_srcname = None,
        b_destname = None,
        b_update = False,
        b_use_syslibs = False,
        b_revision = None,
    ):
    
    if b_revision is None:
        b_revision = 'main'
    # on Linux, rename the binary to `pdfium` to ensure it also works with older versions of ctypesgen
    if b_destname is None and sys.platform.startswith('linux'):
        b_destname = 'pdfium'
    
    if sys.platform.startswith('win32'):
        os.environ['DEPOT_TOOLS_WIN_TOOLCHAIN'] = "0"
    
    depot_dl_done = dl_depottools(b_update)
    if depot_dl_done:
        patch_depottools()
    
    GClient = _get_tool('gclient', 'bat')
    GN      = _get_tool('gn', 'bat')
    Ninja   = _get_tool('ninja', 'exe')
    
    pdfium_dl_done = dl_pdfium(b_update, b_revision, GClient)
    if pdfium_dl_done:
        patch_pdfium()
    
    config_dict = DefaultConfig.copy()
    if b_use_syslibs:
        config_dict.update(SyslibsConfig)
        run_cmd(["python3", "build/linux/unbundle/replace_gn_files.py", "--system-libraries", "icu"], cwd=PDFiumDir)
    config_str = _serialise_config(config_dict)
    print("\nBuild configuration:\n%s\n" % config_str)
    
    configure(config_str, GN)
    build(Ninja)
    libpath = find_lib(b_srcname)
    pack(libpath, b_destname)


def parse_args(argv):
    
    parser = argparse.ArgumentParser(
        description = "A script to automate building PDFium from source and generating bindings with ctypesgen.",
    )
    
    parser.add_argument(
        '--srcname', '-s',
        help = "Name of the generated PDFium binary file. This script tries to automatically find the binary, which should usually work. If it does not, however, this option may be used to explicitly provide the file name to look for.",
    )
    parser.add_argument(
        '--destname', '-d',
        help = "Rename the binary to a different filename.",
    )
    parser.add_argument(
        '--update', '-u',
        action = 'store_true',
        help = "Update existing PDFium/DepotTools repositories, removing local changes.",
    )
    parser.add_argument(
        '--use-syslibs', '-l',
        action = 'store_true',
        help = "Use system libraries instead of those bundled with PDFium. (Make sure that freetype, lcms2, libjpeg, libopenjpeg2, libpng, zlib and icuuc are installed, and that $PKG_CONFIG_PATH is set correctly (e. g. to `/usr/lib/x86_64-linux-gnu/pkgconfig`)",
    )
    parser.add_argument(
        '--revision', '-r',
        help = "PDFium revision to check out (defaults to main).",
    )
    
    return parser.parse_args(argv)


def main_cli(argv=sys.argv[1:]):
    args = parse_args(argv)
    return main(
        b_srcname = args.srcname,
        b_destname = args.destname,
        b_update = args.update,
        b_use_syslibs = args.use_syslibs,
        b_revision = args.revision,
    )
    

if __name__ == '__main__':
    main_cli()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

from pypdfium2._helpers import (
    opener,
    saver,
    toc,
    pdf_renderer,
    page_renderer,
    text_inserter,
)


class PdfDocument:
    """
    Document class that maps the functional support model to an object-oriented API, unifying the separate helpers.
    
    Parameters:
        input_obj: The file or data to load using :func:`.open_pdf_auto`.
        password: A password to unlock the PDF, if encrypted.
    """
    
    def __init__(
            self,
            input_obj,
            password = None,
        ):
        self._input_obj = input_obj
        self._password = password
        self._pdf, self._ld_data = opener.open_pdf_auto(
            self._input_obj,
            password = self._password,
        )
    
    @property
    def raw(self):
        """ Get the raw PDFium ``FPDF_DOCUMENT`` handle. """
        return self._pdf
    
    def close(self):
        """ Close the document to release allocated memory. This method must be called when done with processing the PDF. """
        return opener.close_pdf(self._pdf, self._ld_data)
    
    def save(self, *args, **kws):
        """ Save the PDF into an output byte buffer (see :func:`.save_pdf`). """
        return saver.save_pdf(self._pdf, *args, **kws)
    
    def get_toc(self, **kws):
        """ Incrementally read the document's table of contents (see :func:`.get_toc`). """
        yield from toc.get_toc(self._pdf, **kws)
    
    def render_page_tobytes(self, index, **kws):
        """ Render a single page to bytes (see :func:`.render_page_tobytes`). """
        return page_renderer.render_page_tobytes(self._pdf, page_index=index, **kws)
    
    def render_pdf_tobytes(self, **kws):
        """ Incrementally render multiple pages to bytes (see :func:`.render_pdf_tobytes`). """
        yield from pdf_renderer.render_pdf_tobytes(self._input_obj, password=self._password, **kws)
    
    def render_page_topil(self, index, **kws):
        """ Render a single page to a :mod:`PIL` image (see :func:`.render_page_topil`). """
        return page_renderer.render_page_topil(self._pdf, page_index=index, **kws)
    
    def render_pdf_topil(self, **kws):
        """ Incrementally render multiple pages to :mod:`PIL` images (see :func:`.render_pdf_topil`). """
        yield from pdf_renderer.render_pdf_topil(self._input_obj, password=self._password, **kws)
    
    def insert_text(self, index, **kws):
        """ Insert text into a page (see :func:`.insert_text`). """
        return text_inserter.insert_text(self._pdf, page_index=index, **kws)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import os
import ctypes
from os.path import join
import pypdfium2 as pdfium
from ..conftest import TestFiles, OutputDir


def test_save_pdf_tobuffer():
    
    pdf, ld_data = pdfium.open_pdf_auto(TestFiles.render)
    pdfium.FPDFPage_Delete(pdf, ctypes.c_int(0))
    
    buffer = io.BytesIO()
    pdfium.save_pdf(pdf, buffer, version=17)
    buffer.seek(0)
    
    data = buffer.read()
    
    exp_start = b"%PDF-1.7"
    exp_end = b"%EOF\r\n"
    
    assert data[:len(exp_start)] == exp_start
    assert data[-len(exp_end):] == exp_end
    
    pdfium.close_pdf(pdf, ld_data)


def test_save_pdf_tofile():
    
    src_pdf, ld_data = pdfium.open_pdf_auto(TestFiles.multipage)
    
    # page tiling (n-up)
    dest_pdf = pdfium.FPDF_ImportNPagesToOne(
        src_pdf,
        ctypes.c_float(1190),  # width
        ctypes.c_float(1684),  # height
        ctypes.c_size_t(2),    # number of horizontal pages
        ctypes.c_size_t(2),    # number of vertical pages
    )
    
    output_path = join(OutputDir,'n-up.pdf')
    with open(output_path, 'wb') as file_handle:
        pdfium.save_pdf(dest_pdf, file_handle)
    
    pdfium.close_pdf(src_pdf, ld_data)
    pdfium.close_pdf(dest_pdf)
    
    assert os.path.isfile(output_path)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import configparser
import pkg_resources
from os.path import join, isdir, isfile
from pypdfium2 import V_PYPDFIUM2, V_LIBPDFIUM
from .conftest import *


def test_testpaths():
    for dirpath in (TestDir, SourceTree, ResourceDir, OutputDir):
        assert isdir(dirpath)
    for filepath in iterate_testfiles(False):
        assert isfile(filepath)


def test_versions():
    assert V_PYPDFIUM2 == pkg_resources.get_distribution('pypdfium2').version
    assert V_LIBPDFIUM > 5000 


def test_entrypoint():    
    
    setup_cfg = configparser.ConfigParser()
    setup_cfg.read( join(SourceTree,'setup.cfg') )
    console_scripts = setup_cfg['options.entry_points']['console_scripts']
    
    entry_point = console_scripts.split('=')[-1].strip().split(':')
    module_path = entry_point[0]
    method_name = entry_point[1]
    
    namespace = {}
    exec("from %s import %s" % (module_path, method_name), namespace)
    assert method_name in namespace
    
    function = namespace[method_name]
    assert callable(function)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import shutil
import tempfile
import subprocess
import functools
import urllib.request
from pathlib import Path
from os.path import join, isfile
from concurrent.futures import ThreadPoolExecutor
import pytest
from ..conftest import SourceTree, pl_names
from pl_setup import packaging_base as pkg_base


def test_libnames():
    for name in pkg_base.Libnames:
        assert "pdfium" in name

def test_platformnames():
    for member in pl_names:
        assert member == getattr(pkg_base.PlatformNames, member)

def test_version_namespace():
    vars = ("V_MAJOR", "V_MINOR", "V_PATCH", "V_BETA", "V_PYPDFIUM2", "V_LIBPDFIUM")
    for var in vars:
        assert var in pkg_base.VerNamespace
        assert isinstance(pkg_base.VerNamespace[var], (str, int, type(None)))

def test_paths():
    assert pkg_base.HomeDir == str( Path.home() )
    assert pkg_base.SourceTree == SourceTree
    assert pkg_base.DataTree == str( Path(SourceTree) / "data" )
    assert pkg_base.SB_Dir == str( Path(SourceTree) / "sourcebuild" )
    assert pkg_base.ModuleDir == str( Path(SourceTree) / "src" / "pypdfium2" )
    assert pkg_base.VersionFile == str( Path(pkg_base.ModuleDir) / "_version.py" )


Git = shutil.which("git")

@pytest.mark.skipif(not Git, reason="git not installed")
def test_run_cmd():
    comp_process = pkg_base.run_cmd([Git, "--version"], cwd=None)
    assert isinstance(comp_process, subprocess.CompletedProcess)
    message = pkg_base.run_cmd([Git, "status"], cwd=SourceTree, capture=True)
    assert isinstance(message, str)


def _get_header(url_prefix, include_dir, filename):
    return urllib.request.urlretrieve(url_prefix+filename, join(include_dir.name, filename))

@pytest.mark.skipif(not shutil.which("ctypesgen"), reason="ctypesgen not installed")
def test_call_ctypesgen():
    
    plat_dir = tempfile.TemporaryDirectory()
    include_dir = tempfile.TemporaryDirectory()
    
    header_files = ("fpdfview.h", "fpdf_searchex.h", "fpdf_catalog.h")
    url_prefix = "https://raw.githubusercontent.com/chromium/pdfium/main/public/"
    with ThreadPoolExecutor( len(header_files) ) as pool:
        for rc in pool.map(functools.partial(_get_header, url_prefix, include_dir), header_files):
            pass
    
    pkg_base.call_ctypesgen(plat_dir.name, include_dir.name)
    bindings_file = join(plat_dir.name, "_pypdfium.py")
    assert isfile(bindings_file)
    
    with open(bindings_file, "r") as fh:
        text = fh.read()
        assert "FPDFCatalog_IsTagged" in text
        assert "FPDFText_GetCharIndexFromTextIndex" in text
    
    plat_dir.cleanup()
    include_dir.cleanup()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import logging
from pytest import approx
import pypdfium2 as pdfium
from pypdfium2._cli.toc import print_toc
from ..conftest import TestFiles


def _compare_bookmark(bookmark, title, page_index, view_mode, view_pos, is_closed):
    assert bookmark.title == title
    assert bookmark.page_index == page_index
    assert bookmark.view_mode == view_mode
    assert approx(bookmark.view_pos, rel=0.01) == view_pos
    assert bookmark.is_closed == is_closed


def test_gettoc():
    
    doc = pdfium.PdfDocument(TestFiles.bookmarks)
    toc = doc.get_toc()
    
    # check first bookmark
    _compare_bookmark(
        next(toc),
        title = "One",
        page_index = 0,
        view_mode = pdfium.ViewMode.XYZ,
        view_pos = (89, 758, 0),
        is_closed = True,
    )
    
    # check common values
    for bookmark in toc:
        assert isinstance(bookmark, pdfium.OutlineItem)
        assert bookmark.view_mode is pdfium.ViewMode.XYZ
        assert round(bookmark.view_pos[0]) == 89
    
    # check last bookmark
    _compare_bookmark(
        bookmark,
        title = "Three-B",
        page_index = 1,
        view_mode = pdfium.ViewMode.XYZ,
        view_pos = (89, 657, 0),
        is_closed = False
    )
    
    doc.close()


def test_gettoc_circular(caplog):
    with caplog.at_level(logging.CRITICAL):
        with pdfium.PdfContext(TestFiles.bookmarks_circular) as pdf:
            toc = pdfium.get_toc(pdf)
            print_toc(toc)
            assert "circular bookmark reference" in caplog.text


def test_printtoc():
    with pdfium.PdfContext(TestFiles.bookmarks) as pdf:
        toc = pdfium.get_toc(pdf)
        print()
        print_toc(toc)


def test_gettoc_maxdepth():
    
    doc = pdfium.PdfDocument(TestFiles.bookmarks)
    toc = [b for b in doc.get_toc(max_depth=1)]
    assert len(toc) == 3
    print_toc(toc)
    
    doc.close()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

V_MAJOR = 1
V_MINOR = 9
V_PATCH = 1
V_BETA = None

#: pypdfium2 version string
V_PYPDFIUM2 = "%s.%s.%s" % (V_MAJOR, V_MINOR, V_PATCH)
if V_BETA is not None:
    V_PYPDFIUM2 += "b%s" % V_BETA

#: PDFium library version integer (git tag)
V_LIBPDFIUM = 5065



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

import ctypes
from os.path import abspath
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers.error_handler import handle_pdfium_error


def is_buffer(obj):
    if all( callable(getattr(obj, a, None)) for a in ('seek', 'tell', 'readinto') ):
        return True
    else:
        return False


class _reader_class:
    
    def __init__(self, buffer):
        self.buffer = buffer
    
    def __call__(self, param, position, p_buf, size):
        c_buf = (ctypes.c_char * size).from_address( ctypes.addressof(p_buf.contents) )
        self.buffer.seek(position)
        self.buffer.readinto(c_buf)
        return 1


class LoaderDataHolder:
    
    def __init__(self, reader_func, buffer, autoclose):
        self.reader_func = reader_func
        self.buffer = buffer
        self.autoclose = autoclose
    
    def close(self):
        # access the reader variable again to make sure that even a heavily optimising interpreter would not prematurely delete the object
        id(self.reader_func)
        if self.autoclose:
            self.buffer.close()


def open_pdf_buffer(buffer, password=None, autoclose=False):
    
    if not is_buffer(buffer):
        raise ValueError("Buffer must implement the methods seek(), tell(), and readinto().")
    
    buffer.seek(0, 2)
    file_len = buffer.tell()
    buffer.seek(0)
    
    FuncType = ctypes.CFUNCTYPE(
        # restype
        ctypes.c_int,
        # argtypes
        ctypes.POINTER(None),
        ctypes.c_ulong,
        ctypes.POINTER(ctypes.c_ubyte),
        ctypes.c_ulong,
    )
    
    fileaccess = pdfium.FPDF_FILEACCESS()
    fileaccess.m_FileLen = file_len
    fileaccess.m_GetBlock = FuncType( _reader_class(buffer) )
    
    pdf = pdfium.FPDF_LoadCustomDocument(ctypes.byref(fileaccess), password)
    ld_data = LoaderDataHolder(fileaccess.m_GetBlock, buffer, autoclose)
    
    if pdfium.FPDF_GetPageCount(pdf) < 1:
        handle_pdfium_error(False)
    
    return pdf, ld_data


def open_pdf_native(filepath, password=None):
    file_handle = open(abspath(filepath), "rb")
    return open_pdf_buffer(file_handle, password=password, autoclose=True)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import functools
import concurrent.futures
from pypdfium2 import _pypdfium as pdfium
from pypdfium2._helpers import page_renderer
from pypdfium2._helpers.opener import PdfContext


def _process_page(index, render_meth, input_obj, password, **kws):
    with PdfContext(input_obj, password) as pdf:
        result = render_meth(pdf, index, **kws)
    return index, result


def render_pdf_base(
        render_meth,
        input_obj,
        page_indices = None,
        password = None,
        n_processes = os.cpu_count(),
        **kws
    ):
    """
    Rasterise multiple pages of a PDF using an arbitrary page rendering method.
    Base function for :func:`.render_pdf_tobytes` and :func:`.render_pdf_topil`.
    
    Parameters:
        input_obj (str | bytes | typing.BinaryIO):
            The PDF document to render. It may be given as file path, bytes, or byte buffer.
        page_indices (typing.Sequence[int]):
            A list of zero-based indices of the pages to render.
        password (typing.Optional[str | bytes]):
            A password to unlock the document, if encrypted.
        n_processes (int):
            The number of rendering processes to run in parallel (defaults to :func:`os.cpu_count`).
        kws (dict):
            Additional parameters to be passed to :func:`.render_page_base`.
    """
    
    with PdfContext(input_obj, password) as pdf:
        n_pages = pdfium.FPDF_GetPageCount(pdf)
    
    if page_indices is None or len(page_indices) == 0:
        page_indices = [i for i in range(n_pages)]
    if not all(0 <= i < n_pages for i in page_indices):
        raise ValueError("Out of range page index detected.")
    
    n_digits = len(str( max(page_indices)+1 ))
    invoke_renderer = functools.partial(_process_page, render_meth=render_meth, input_obj=input_obj, password=password, **kws)
    
    with concurrent.futures.ProcessPoolExecutor(n_processes) as pool:
        for index, image in pool.map(invoke_renderer, page_indices):
            suffix = str(index+1).zfill(n_digits)
            yield image, suffix


def render_pdf_tobytes(*args, **kws):
    """
    Render multiple pages of a PDF to bytes. See :func:`.render_pdf_base` and :func:`.render_page_tobytes`.
    
    Yields:
        :class:`tuple`, :class:`str`
    """
    yield from render_pdf_base(page_renderer.render_page_tobytes, *args, **kws)


def render_pdf_topil(*args, **kws):
    """
    Render multiple pages of a PDF to :mod:`PIL` images. See :func:`.render_pdf_base` and :func:`.render_page_topil`.
    
    Yields:
        :class:`PIL.Image.Image`, :class:`str`
    """
    yield from render_pdf_base(page_renderer.render_page_topil, *args, **kws)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from pypdfium2 import _pypdfium as pdfium


class PdfiumError (RuntimeError):
    """ An exception from the PDFium library. """
    pass


def handle_pdfium_error(valid=True):
    """
    Check the last PDFium error code and raise an exception accordingly.
    
    Parameters:
        valid (bool):
            If :data:`False`, also raise an exception if ``FPDF_GetLastError()`` returns ``FPDF_ERR_SUCCESS``.
    Returns:
        :class:`int` – The error code as returned by PDFium.
    """
    
    last_error = pdfium.FPDF_GetLastError()
    
    if last_error == pdfium.FPDF_ERR_SUCCESS:
        if not valid:
            raise PdfiumError("Even though no errors were reported, something invalid happened.")
    elif last_error == pdfium.FPDF_ERR_UNKNOWN:
        raise PdfiumError("An unknown error occurred.")
    elif last_error == pdfium.FPDF_ERR_FILE:
        raise PdfiumError("The file could not be found or opened.")
    elif last_error == pdfium.FPDF_ERR_FORMAT:
        raise PdfiumError("Data format error.")
    elif last_error == pdfium.FPDF_ERR_PASSWORD:
        raise PdfiumError("Missing or wrong password.")
    elif last_error == pdfium.FPDF_ERR_SECURITY:
        raise PdfiumError("Unsupported security scheme.")
    elif last_error == pdfium.FPDF_ERR_PAGE:
        raise PdfiumError("Page not found or content error.")
    else:
        raise ValueError("Unknown PDFium error code %s" % last_error)
    
    return last_error



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import shutil
import tempfile
from os.path import join
from pl_setup import update_pdfium as fpdf_up
from pl_setup.packaging_base import (
    PlatformNames,
    VersionFile,
    VerNamespace,
    _get_ver_namespace,
)
from ..conftest import pl_names


def test_consts():
    assert fpdf_up.ReleaseRepo == "https://github.com/bblanchon/pdfium-binaries"
    assert fpdf_up.ReleaseURL == "https://github.com/bblanchon/pdfium-binaries/releases/download/chromium%2F"
    assert fpdf_up.ReleaseExtension == "tgz"


def test_releasenames():
    assert len(fpdf_up.ReleaseNames) == len(pl_names) - 1
    for key, value in fpdf_up.ReleaseNames.items():
        assert hasattr(PlatformNames, key)
        prefix, system, cpu = value.replace("linux-musl", "musllinux").split("-", maxsplit=3)
        assert prefix == "pdfium"
        assert system in ("linux", "musllinux", "mac", "win")
        assert cpu in ("x64", "x86", "arm64", "arm")


def test_setversion():
    
    tempdir = tempfile.TemporaryDirectory()
    tmp_versionfile = join(tempdir.name, "tmp_versionfile.py")
    shutil.copy(VersionFile, tmp_versionfile)
    
    target_version = VerNamespace["V_LIBPDFIUM"] + 10
    fpdf_up.handle_versions(tmp_versionfile, target_version)
    new_vernsp = _get_ver_namespace(tmp_versionfile)
    
    assert new_vernsp["V_LIBPDFIUM"] == target_version
    assert new_vernsp["V_MINOR"] == VerNamespace["V_MINOR"] + 1
    
    tempdir.cleanup()


def test_getversion():
    latest = fpdf_up.get_latest_version()
    assert latest > 5000
    assert latest >= VerNamespace["V_LIBPDFIUM"]



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

from ctypes import c_float, byref
from pypdfium2 import _pypdfium as pdfium


def _get_box(page, box_function, fallback_function):
    
    left, bottom, right, top = c_float(), c_float(), c_float(), c_float()
    
    ret_code = box_function(page, byref(left), byref(bottom), byref(right), byref(top))
    if not ret_code:
        return fallback_function(page)
    
    return (left.value, bottom.value, right.value, top.value)


def get_mediabox(page):
    """
    Get the MediaBox of *page* in PDF canvas units (usually 1/72in).
    Falls back to ANSI A (0, 0, 612, 792) if the page does not define a MediaBox.
    
    Parameters:
        page (``FPDF_PAGE``): PDFium page object handle.
    Returns:
        A tuple of four float coordinates.
    """
    return _get_box(page, pdfium.FPDFPage_GetMediaBox, lambda _p: (0, 0, 612, 792))


def get_cropbox(page):
    """ Get the CropBox of *page* (Fallback: :func:`get_mediabox`). """
    return _get_box(page, pdfium.FPDFPage_GetCropBox, get_mediabox)

def get_bleedbox(page):
    """ Get the BleedBox of *page* (Fallback: :func:`get_cropbox`). """
    return _get_box(page, pdfium.FPDFPage_GetBleedBox, get_cropbox)

def get_trimbox(page):
    """ Get the TrimBox of *page* (Fallback: :func:`get_cropbox`). """
    return _get_box(page, pdfium.FPDFPage_GetTrimBox, get_cropbox)

def get_artbox(page):
    """ Get the ArtBox of *page* (Fallback: :func:`get_cropbox`). """
    return _get_box(page, pdfium.FPDFPage_GetArtBox, get_cropbox)



# SPDX-FileCopyrightText: 2021 Adam Huganir <adam@huganir.com>
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pytest
import os.path
import pypdfium2 as pdfium
from pypdfium2._cli import main
from pypdfium2._cli import render


def test_rotation_type():
    test_cases = [
        ("0", 0),
        ("90", 90),
        ("180", 180),
        ("270", 270),
    ]
    for input, expectation in test_cases:
        assert render.rotation_type(input) == expectation


def test_rotation_type_fail_oob():
    with pytest.raises(ValueError, match="Invalid rotation value"):
        render.rotation_type("101")
    with pytest.raises(ValueError, match="invalid literal for int()"):
        render.rotation_type("string")


def test_colour_type():
    test_cases = [
        ("None", None),
        ("none", None),
    ]
    for input, expectation in test_cases:
        assert render.colour_type(input) == expectation


def test_pagetext_type():
    test_cases = [
        ("", None),
        ("1,2", [0, 1]),
        ("3-5", [2, 3, 4]),
        ("5-3", [4, 3, 2]),
        ("1", [0]),
    ]
    for input, expectation in test_cases:
        assert render.pagetext_type(input) == expectation


def test_parse_args():
    
    argv = [
        'render',
        'path/to/document.pdf',
        '-o', 'output_dir/',
        '--pages', '1,4,5-7,6-4',
        '--scale', '2',
        '--rotation', '90',
        '--colour', '(255, 255, 255, 255)',
        '--optimise-mode', 'none',
        '--processes', '4',
    ]
    
    args = main.parse_args(argv)
    
    assert args.inputs == ['path/to/document.pdf']
    assert args.output == os.path.abspath('output_dir/')
    assert args.pages == [0, 3, 4, 5, 6, 5, 4, 3]
    assert args.scale == 2
    assert args.rotation == 90
    assert args.colour == (255, 255, 255, 255)
    assert args.no_annotations == False
    assert args.optimise_mode == pdfium.OptimiseMode.none
    assert args.greyscale == False
    assert args.processes == 4



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# full support model namespace

from pypdfium2._helpers.constants import *
from pypdfium2._helpers.error_handler import *
from pypdfium2._helpers.utilities import *
from pypdfium2._helpers.opener import *
from pypdfium2._helpers.nativeopener import *
from pypdfium2._helpers.page_renderer import *
from pypdfium2._helpers.pdf_renderer import *
from pypdfium2._helpers.toc import *
from pypdfium2._helpers.saver import *
from pypdfium2._helpers.boxes import *
from pypdfium2._helpers.classes import *
from pypdfium2._helpers.text_inserter import *



# PYTHON_ARGCOMPLETE_OK
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import argparse
from pypdfium2._version import (
    V_PYPDFIUM2,
    V_LIBPDFIUM,
)
from pypdfium2._cli import (
    render,
    toc,
    merge,
    tile,
)

try:
    import argcomplete
except ImportError:
    have_argcomplete = False
else:
    have_argcomplete = True


Subcommands = dict(
    render = render,
    toc = toc,
    merge = merge,
    tile = tile,
)


def parse_args(argv=sys.argv[1:]):
    
    parser = argparse.ArgumentParser(
        prog = "pypdfium2",
        description = "Command line interface to the pypdfium2 Python library",
    )
    parser.add_argument(
        "--version", "-v",
        action = "version",
        version = "pypdfium2 %s (libpdfium %s)" % (V_PYPDFIUM2, V_LIBPDFIUM),
    )
    
    subparsers = parser.add_subparsers(dest="subcommand")
    for cmd in Subcommands.values():
        cmd.attach_parser(subparsers)
    
    if have_argcomplete:
        argcomplete.autocomplete(parser)
    
    return parser.parse_args(argv)


def main():
    args = parse_args()
    if not args.subcommand:
        print("One of the following subcommands must be given: %s" % [sc for sc in Subcommands.keys()])
        return
    Subcommands[args.subcommand].main(args)
